'''
Class based templates
=====================

This minimalistic template system utilizes python class hierarchy, class attributes, and standard string formatting.

The basics
----------

Templates are defined as subclasses of the base ``clabate.Template`` class::

    class MyTemplate(clabate.Template):

        sentence = 'This is {article} {fruit_name}. The {fruit_name} is {fruit_attribute}.'

        # some defaults
        fruit_name = 'apple'
        fruit_attribute = 'green'

        # Template properties can be used to build values dynamically.
        # Such properties have ``context`` argument which provides access to the rendering context:
        @property
        def article(self, context=None):
            # this is a simplified logic just for illistration
            if context['fruit_name'].lower()[0] in 'bcdfghjklmnpqrstuvwxyz':
                return 'a'
            else:
                return 'an'

    template = MyTemplate()
    print(template.render())

The result of rendering::

    {'article': 'an', 'fruit_name': 'apple', 'fruit_attribute': 'green', 'sentence': 'This is an apple. The apple is green.'}

The data for the template can be provided at instantiation time and/or at rendering time::

    template = MyTemplate(fruit_name='banana')
    print(template.render(fruit_attribute='yellow'))

This produces the following result::

    {'fruit_name': 'banana', 'fruit_attribute': 'green', 'article': 'a', 'sentence': 'This is a banana. The banana is yellow.'}


What is formatted
-----------------

Class attributes of ``str`` type are always formatted, regardless if they contain substitutions or not.
Thus, braces must be always escaped.

XXX Strings are formatted by calling ``.format_map`` method.

``string.Template`` is not supported. Previously it was supported but
later on it was dropped for sake of simplicity.

C-style formatting is not supported.
Although the initial idea was to support both styles,
C-style was rejected because of some issues with detecting dependencies.

If a string does not contain any substitutions, it is not formatted at all.
In the above example ``fruit_name`` and ``fruit_attribute`` are not formatted.

Values returned by template properties are not formatted. --- XXX revise this, probably use some decorator?

Rendering context contains all parameters for templates and is populated with formatted values.


Attribute names
---------------

Attributes that start from underscore are treated as private and excluded from processing.
Underscores are allowed for values passed to ``render``.

Escaping
--------

Everything is escaped.

In ``MarkupTemplate`` use ``Markup`` instead of ``str`` to avoid escaping.

Escaping or tag attributes can be done better with ???
XXX: this will be dropped???: ``TagAttr`` wrapper for better escaping of attribute values.

Avoiding double escaping ???


Thread-safety
-------------

Rendering methods do not modify object state so the template can be instantiated once, and rendered by multiple threads.


HTML templates example
----------------------

Let's define basic HTML structure::

    class BaseTemplate(clabate.MarkupPackerMixIn, clabate.MarkupTemplate):

        lang = "en"

        html = """
            <!DOCTYPE html>
            <html lang="{lang}">
            <head>
                <meta charset="{encoding}">
                <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
                <link rel="icon" type="image/svg+xml" href="/favicon.svg">
                <title>{title}</title>
                {stylesheets}
                {scripts}
            </head>
            <body>
                <header>
                    {header}
                </header>
                <main>
                    {body}
                </main>
                <footer>
                    {footer}
                </footer>
            </body>
            </html>
        """

        title = ""
        stylesheets = ""
        scripts = ""
        body = ""
        header = ""
        footer = ""

Then, define the main page::

    class MainPage(BaseTemplate):

        title = "Welcome to Clabate"

        body = """
            <div class="thumbnails">
                {thumbnails}
            </div>
        """

        @property
        def thumbnails(self, context=None):
            resources = [
                ('page_uri_1', 'thumbnail_uri_1', 'resource_uri_1'),
                ('page_uri_2', 'thumbnail_uri_2', 'resource_uri_2')
            ]
            for page_uri, thumbnail_uri, resource_uri in resources:
                yield self.render(
                    '<a href={page_uri}><img src={thumbnail_uri}></a>',
                    page_uri = clabate.TagAttr(page_uri),
                    thumbnail_uri = clabate.TagAttr(thumbnail_uri)
                )

:copyright: Copyright 2019-2022 AXY axy@declassed.art
:license: BSD, see LICENSE for details.
'''

# TODO fix docstrings style

__version__ = '0.3.5'

# shortcut imports
from .core import Template, Comment, CommentMixin
from .markup import MarkupTemplate, Markup, TagAttr, MinifiedMarkupTemplate, minified
