'''
Implementation of the basic Template class

:copyright: Copyright 2019-2021 AXY axy@declassed.art
:license: BSD, see LICENSE for details.
'''

from collections import defaultdict
import inspect
import string
import textwrap
import types

#-----------------------------------------------------------------------------------
# Escaping layer.
#
# Substitution values should be escaped only once.
# This implementation uses ids.
# We could use subclasses to track already escaped values,
# but that's not a good way either.

class EscapeContext(dict):
    '''
    Base rendering context class that tracks escaped strings to avoid double escaping.
    Escaped strings are tracked by their ids, returned by built-in `id()` function.
    '''
    def __init__(self, template):
        self.__template__ = template
        self.__escaped__ = set()

    def set_escaped_item(self, key, value):
        '''
        Special method to set escaped value.
        Such values are generated by EscapeFormatter.
        '''
        self[key] = value
        if isinstance(value, str):
            self.__escaped__.add(id(value))

    def is_escaped(self, value):
        '''
        Check if a particular value is escaped.
        '''
        return id(value) in self.__escaped__

    def __delitem__(self, key):
        # discard escaped mark, if any
        self.__escaped__.discard(id(self[key]))
        super().__delitem__(key)

class EscapeFormatter(string.Formatter):
    '''
    EscapeFormatter calls ``Template.escape`` method for substitution values
    and literal text.
    Literal text is always escaped.
    Substitution values should be escaped only once, so the `format_field` method
    calls `is_escaped` method of `EscapeContext`.
    '''
    def __init__(self, context):
        '''
        Initialize formatter with context object.
        We have to provide context at initilization stage because we need
        to check escaped values with ``is_escaped`` method.
        '''
        super().__init__()
        self.__context__ = context

    def xformat(self, format_string):
        '''
        Wrapper for vformat method.
        Preserve type of format_string, which can be a subclass of str, i.e. Markup.
        '''
        format_string_class = type(format_string)
        value = super().vformat(format_string, tuple(), self.__context__)
        # preserve type of format_string
        if format_string_class is not str:
            value = format_string_class(value)
        return value

    def parse(self, format_string):
        '''
        Wrapper to escape literal text.
        '''
        # Template.escape relies on the type of value to escape, i.e. it could be str, Markup, etc.
        # So if the type of format string is not ``str``
        # we need to convert pieces of literal text to this type before escaping.
        format_string_class = type(format_string)
        convert_class = format_string_class is not str
        # Parse format string.
        for literal_text, field_name, format_spec, conversion in super().parse(format_string):
            # Convert the type of literal text and format_spec.
            if convert_class:
                if literal_text:
                    literal_text = format_string_class(literal_text)
                if format_spec:
                    format_spec = format_string_class(format_spec)
            if literal_text:
                # Escape literal text. Don't escape format_spec because this will be processed recursively.
                literal_text = self.__context__.__template__.escape(literal_text)
            yield literal_text, field_name, format_spec, conversion

    def format_field(self, value, format_spec):
        '''
        Wrapper to escape substitutions except those that shouldn't be escaped.
        Check them with ``is_escaped`` method of context object.
        '''
        value = super().format_field(value, format_spec)
        if not self.__context__.is_escaped(value):
            value = self.__context__.__template__.escape(value)
        return value

#-----------------------------------------------------------------------------------
# Rendering context.

class Context(EscapeContext):
    '''
    Rendering context for templates.

    This sublass implements the following:
    1. Call property getters and cache returned values.
    2. Handle missing values by calling template's ``missing`` method.
    '''
    def __init__(self, template):
        super().__init__(template)
        self.__missing_types__ = dict()

    def __missing__(self, name):
        '''
        Try calling property getter, if missing, call template's ``missing`` method.
        '''
        if name in self.__template__._properties:
            return self.call_property_getter(name)
        else:
            return self.get_missing_value(name)

    def call_property_getter(self, name):
        '''
        Call ``fget`` method of property
        Save returned value.
        '''
        value = self.__template__._properties[name].fget(self.__template__, self)
        # escape value
        if isinstance(value, str):
            value = self.__template__.escape(value)
        # save value
        self.set_escaped_item(name, value)
        return value

    def get_missing_value(self, name):
        '''
        Get substitution for missing value.
        Save returned value.
        '''
        value = self.__template__.missing(name)
        # escape value
        if isinstance(value, str):
            value = self.__template__.escape(value)
            # Preserve subclass of value which can be Markup, for example.
            value = self.make_missing_value(value)
        # save value
        self.set_escaped_item(name, value)
        return value

    def make_missing_value(self, value):
        '''
        Use MissingValue as a mixin to override necessary methods of value class.

        Args:
            value: an instance of ``str`` or a subclass of ``str``

        Returns:
            An instance of type(value) and MissingValue.
        '''
        bases = (MissingValue, type(value))
        if bases in self.__missing_types__:
            value_type = self.__missing_types__[bases]
        else:
            value_type = type('MissingValue', bases, {})
            self.__missing_types__[bases] = value_type
        return value_type(value)

class MissingValue:
    '''
    Mixin class for missing substitutions.
    For simple substitutions this mixin is unnecessary but for
    more or less complex ones such as

        {foo.bar[1][value]}

    it helps to identify missing things.
    Otherwise, the traceback would be somewhat cryptic:

        File "/usr/lib/python3.7/string.py", line 301, in get_field
            obj = getattr(obj, i)
        AttributeError: 'str' object has no attribute 'bar'
    '''
    def __getattr__(self, name):
        return self

    def __getitem__(self, name):
        return self

    def __format__(self, format_spec):
        '''
        This method is required when the missing substitution contains format specification.
        '''
        return self

#-----------------------------------------------------------------------------------
# Base template class

class Template:
    '''
    Base template class.
    '''

    _formatter_class = EscapeFormatter

    def __init__(self, **kwargs):
        # Set object attributes from kwargs.
        for k, v in kwargs.items():
            setattr(self, k, v)

        class_attributes = self._collect_class_attributes()

        # Group class atributes into:
        self._data_attributes = dict()  # object attributes other than strings
        self._properties = dict()       # object methods decorated with @property
        potential_templates = dict()    # object attributes of string types

        for k, v in class_attributes.items():
            if isinstance(v, str):
                # Attributes of ``str`` type are potential templates.
                potential_templates[k] = v
            else:
                # All other attributes are template attributes or template properties.
                if self._is_template_property(v):
                    self._properties[k] = v
                else:
                    self._data_attributes[k] = v

        # Find real templates and their dependencies.
        template_dependencies = self._collect_template_dependencies(potential_templates)

        # Determine formatting order.
        format_order = self._determine_formatting_order(template_dependencies)

        # Build ``_template_attributes`` list.
        # This list contain templates in formatting order. Each item is a tuple: (attribute name, template)
        self._template_attributes = []
        for attr_name in format_order:
            self._template_attributes.append((attr_name, potential_templates[attr_name]))
            del potential_templates[attr_name]

        # All the rest strings in ``potential_templates`` are not templates.
        # Format and append them to ``_data_attributes``.
        # Always format all these strings because they may contain escaped braces {{...}} and no substitutions.
        # Also, formatting performs escaping.
        formatter = self._formatter_class(Context(self))
        for attr_name, attr_value in potential_templates.items():
            self._data_attributes[attr_name] = formatter.xformat(attr_value)

    def _is_template_property(self, attr):
        '''
        Check if ``attr`` is a special property with additional ``context`` argument.
        '''
        if isinstance(attr, property):
            return 'context' in inspect.signature(attr.fget).parameters
        else:
            return False

    def _collect_class_attributes(self):
        '''
        Collect class attributes including all base classes
        in the hierarchy except object and this, Template class.
        '''
        class_attributes = dict()
        # Reverse method resolution order and exclude first two entries: base ``object`` class and self:
        reverse_mro = inspect.getmro(self.__class__)[-2::-1]
        for template_class in reverse_mro:
            for k, v in vars(template_class).items():
                # Exclude private attributes.
                if k.startswith('_'):
                    continue
                # Exclude methods
                if isinstance(v, types.FunctionType):
                    continue
                class_attributes[k] = v
        return class_attributes

    def _collect_template_dependencies(self, potential_templates):
        '''
        Try running formatting with dummy data.
        This detects real templates and their dependencies.

        Args:
            potential_templates (dict): attribute names and their values.

        Returns:
            dict: keys are template attribute names, values are names of the data they want.
        '''
        class DummyValue(int):
            # All formatting directives safely substitute int values.
            # Add a couple of methods to simulate nested data.

            def __getattr__(self, name):
                return DummyValue(0)

            def __getitem__(self, key):
                return DummyValue(0)

            def __format__(self, format_spec):
                return ''

        class MissingDataCollector(Context):

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.__collected_attributes__ = set()

            def __missing__(self, name):
                self.__collected_attributes__.add(name)
                return DummyValue(0)

        template_dependencies = defaultdict(set)
        for template_attr_name, template_string in potential_templates.items():
            template_data = MissingDataCollector(self)
            formatter = self._formatter_class(template_data)
            formatter.xformat(template_string)
            for other in template_data.__collected_attributes__:
                template_dependencies[template_attr_name].add(other)
        return template_dependencies

    def _determine_formatting_order(self, template_dependencies):
        '''
        Find out formatting order for template strings.
        '''
        # Build dependency graph.
        dependency_graph = defaultdict(list)
        for template_attr_name, others in template_dependencies.items():
            for other in others:
                dependency_graph[template_attr_name].append(other)
                dependency_graph[other] = dependency_graph[other]

        # Topological sort.
        # Cycles are not detected!
        format_order = []

        visited = dict((node_name, False) for node_name in dependency_graph)

        def sort_visit(node_name):
            visited[node_name] = True
            for other in dependency_graph[node_name]:
                if not visited[other]:
                    sort_visit(other)
            format_order.append(node_name)

        for node_name in dependency_graph:
            if not visited[node_name]:
                sort_visit(node_name)

        # Filter format order by real templates.
        return list(filter(lambda name: name in template_dependencies, format_order))

    def render(self, **kwargs):
        '''
        Render self.

        Collect all properties and attributes into a rendering context dict and then
        format all template strings defined as class attributes.

        Returns:
            dict: entire rendering context containing formatted class attributes
        '''
        context = self._prepare_rendering_context(**kwargs)
        return self.render_in_context(context, **kwargs)

    def render_in_context(self, context, **kwargs):
        '''
        Render self in existing `context`.
        '''
        formatter = self._formatter_class(context)
        # Format class attribute strings.
        for attr_name, template_string in self._template_attributes:
            formatted_value = formatter.xformat(template_string)
            # Call post-process method.
            formatted_value = self.post_process(formatted_value)
            # Set formatted value to the rendering context. The value is escaped.
            context.set_escaped_item(attr_name, formatted_value)
        # Return rendering context containing formatted strings.
        return context

    def render_str(self, s, **kwargs):
        '''
        Render template string.

        Collect all properties and attributes into a rendering context dict and then
        format provided template string.

        Returns:
            str: formatted template string
        '''
        context = self._prepare_rendering_context(**kwargs)
        return self.render_str_in_context(s, context, **kwargs)

    def render_str_in_context(self, s, context, **kwargs):
        '''
        Render template string in existing `context`.
        '''
        formatter = self._formatter_class(context)
        # Format provided template string.
        result = formatter.xformat(s)
        # Call post-process method.
        result = self.post_process(result)
        return result

    def _prepare_rendering_context(self, **kwargs):
        '''
        Prepare rendering context.

        Returns:
            an instance of Context class
        '''
        context = Context(self)

        # First, populate ``context`` with data attributes.
        # Strings in _data_attributes are already escaped.
        for k, v in self._data_attributes.items():
            if isinstance(v, str):
                context.set_escaped_item(k, v)
            else:
                context[k] = v

        # Second, populate ``context`` with variables passed to ``__init__`` or set in any other way for ``self``:
        for k, v in vars(self).items():
            # exclude private attributes
            if k.startswith('_'):
                continue
            context[k] = v

        # Third, populate ``context`` with kwargs:
        context.update(kwargs)

        return context

    def missing(self, name):
        '''
        Handler for missing template data.

        Args:
            name: name of an attribute, missing from the rendering context.
        '''
        raise AttributeError('%s does not provide %s' % (self.__class__.__name__, name))

    def escape(self, value):
        '''
        Basic escape method does nothing, but subclasses may override it if escaping is necessary.
        '''
        return value

    def post_process(self, formatted_value):
        '''
        Post-process formatted value.
        This basic method does nothing, but subclasses may override it for special processing.
        '''
        return formatted_value

    @classmethod
    def dedent(self, text):
        '''
        dedent helper, removing leading empty lines
        '''
        return textwrap.dedent(text).lstrip('\r\n')

#-----------------------------------------------------------------------------------
# Comment mixin

class Comment():
    '''
    Comment class provides __format__ method to swallow comments.
    '''

    def __format__(self, format_spec):
        '''
        Comment is in format_spec. Simply ignore it and return empty string.
        '''
        return ''

class CommentMixin():
    '''
    Comment mixin for templates.

    Example::

        The red dog
        {comment: or fox?}
        jumps over the lazy fox. {comment: at last, maybe it's
        a fox who jumps over the lazy dog?}
    '''
    comment = Comment()
