'''
Basic examples.

:copyright: Copyright 2021 AXY axy@declassed.art
:license: BSD, see LICENSE for details.
'''

from textwrap import dedent
from types import SimpleNamespace

from .. import Template

class GoodTemplate1(Template):

    '''
    Example 1.
    '''
    snippet = 'This is {article} {fruit_name}. The {fruit_name} is {fruit_attribute}.'

    '''
    Simple static substitutions.
    '''
    fruit_name = 'apple'
    fruit_attribute = 'green'

    '''
    Properties can be used to build values dynamically.
    Unlike regular properties they should have context argument which provides access to the rendering context:
    '''
    @property
    def article(self, context=None):
        if context is None:
            # Clabate does not call property getters without content.
            # Such a case is when the property is accessed in the usual manner.
            # Simply return something.
            return 'a'
        # Simplified logic just for illistration:
        if context['fruit_name'].lower()[0] in 'bcdfghjklmnpqrstuvwxyz':
            return 'a'
        else:
            return 'an'

    result = 'This is an apple. The apple is green.'

class GoodTemplate2(Template):

    '''
    Example 2.
    Substitutions for templates can be provided via:
     * class attributes
     * keyword args when the class is instantiated
     * keyword args when template is rendered
    '''
    snippet = 'How do {who} {verb} {what}?'
    who = 'they'
    '''
    Rendering the template this way

        template = GoodTemplate2(verb='do')
        context = template.render(what='that')

    will produce the following
    '''
    result = 'How do they do that?'

class GoodTemplate3(Template):

    '''
    Example 3.
    Formatting order and braces escaping.
    '''
    snippet = 'This snippet on {substitution1}. {btw}'
    substitution1 = 'the first substitution which, in turn, depends on {substitution2}'
    substitution2 = 'the second substitution which, in turn, depends on {substitution3}'
    substitution3 = 'the third substitution'
    btw = 'By the way, all class attributes of string type are formatted, {so}.'
    so = 'so braces should always be escaped, like this: {{ }}'

    result = 'This snippet on the first substitution which, in turn, depends on the second substitution'\
             ' which, in turn, depends on the third substitution. By the way, all class attributes'\
             ' of string type are formatted, so braces should always be escaped, like this: {{ }}.'

    '''
    Let's look at some internals:
    '''
    def render(self, *args, **kwargs):
        print('Example 3:')
        print('Data attributes:')
        for a in self._data_attributes.items():
            print(a)
        print('Formatting order:')
        for a in self._template_attributes:
            print(a)
        print()
        return super().render(*args, **kwargs)

class GoodTemplate4(Template):

    '''
    Example 4.
    A bit more complex data and missing values:
    '''
    snippet1 = dedent('''
        What is a.b[name].c[1].d? It's {a.b[name].c[1].d}!
        And what is foo.bar? {foo.bar}
        Wait, what is bar[foo][1]? {bar[foo][1]}
    ''')

    a = SimpleNamespace(
        b = dict(
            name = SimpleNamespace(
                c = [
                    0,
                    SimpleNamespace(
                        d = 'wow'
                    )
                ]
            )
        )
    )
    result1 = dedent('''
        What is a.b[name].c[1].d? It's wow!
        And what is foo.bar? foo is missing from the rendering context!
        Wait, what is bar[foo][1]? bar is missing from the rendering context!
    ''')

    def missing(self, name):
        return '%s is missing from the rendering context!' % name

    '''
    Specifically, this is the case where __format__ method is required for MissingValue, see clabate/core.py:
    '''
    snippet2 = 'Missing {{baz:my format spec}}? {baz:my format spec}'
    result2 = 'Missing {{baz:my format spec}}? baz is missing from the rendering context!'

class GoodTemplate5(Template):

    '''
    Example 5.
    '''
    snippet = 'I love {something_to_love} and hate {something_to_hate}'
    '''
    Before any value gets formatted, the rendering context is populated with static values.
    Properties are evaluated lazily. Their getters are called once and values are cached in the rendering context.
    What if you need a value not evaluated yet?
    '''
    @property
    def something_to_love(self, context):
        # if context is not necessary, access the property in the usual manner:
        return '%s freedom' % self.kind_of_freedom

    @property
    def something_to_hate(self, context):
        # otherwise, use the rendering context to invoke property getter via Clabate:
        print('Example 5:')
        print('context has no activist_type yet:', context)
        value = '%s activists' % context['activist_type']
        print('now it has that:', context)
        return value

    '''
    Note that context can be a keyword or positional argument.
    If a property is accessed in the usual manner it should be a keyword argument because
    property decorator calls the getter without any arguments except self.
    If a property is accessed only from templates or indirectly via context,
    this argument can be positional.
    It's recommended to always declare it as a keyword argument.
    '''
    @property
    def kind_of_freedom(self, context=None):
        return 'absolute'

    @property
    def activist_type(self, context):
        return 'open source'

    result = 'I love absolute freedom and hate open source activists'

def test_basic():
    template = GoodTemplate1()
    context = template.render()
    assert context['snippet'].strip() == template.result.strip()

    template = GoodTemplate2(verb='do')
    context = template.render(what='that')
    assert context['snippet'].strip() == template.result.strip()

    template = GoodTemplate3()
    context = template.render()
    assert context['snippet'].strip() == context['result'].strip()  # Use context['result'] because of braces escaping

    template = GoodTemplate4()
    context = template.render()
    assert context['snippet1'].strip() == template.result1.strip()
    assert context['snippet2'].strip() == context['result2'].strip()  # Use context['result2'] because of braces escaping

    template = GoodTemplate5()
    context = template.render()
    assert context['snippet'].strip() == template.result.strip()
