'''
File inclusion examples.

:copyright: Copyright 2021 AXY axy@declassed.art
:license: BSD, see LICENSE for details.
'''

import os
from textwrap import dedent

from .. import MarkupTemplate, Markup
from ..extras.inclusion import FileReaderMixin, BadPathError

class BaseTemplate(
        MarkupTemplate,
        FileReaderMixin(
            'examples',  # attribute name
            os.path.dirname(__file__)
        )
      ):
    pass

class GoodTemplate(BaseTemplate):

    # Example 1.
    # Include self. If you save the output, it should be displayed correctly in the browser.
    snippet1 = Markup(dedent('''
        <h1>Content of file_inclusion.py</h1>
        <pre>
        {examples:read('file_inclusion.py')}
        </pre>
    '''))

    # Example 2.
    # Include sample HTML file for automated testing.
    snippet2 = Markup(dedent('''
        <h1>Content of file_inclusion.html</h1>
        <pre>
        {examples:read('file_inclusion.html')}
        </pre>
    '''))
    # Included file should be properly escaped:
    result2 = dedent('''
        <h1>Content of file_inclusion.html</h1>
        <pre>
        &lt;div&gt;
            Sample &lt;strong&gt;inclusion&lt;/strong&gt; file.
        &lt;/div&gt;

        </pre>
    ''')

    # Example 3.
    # Include sample HTML file as markup for automated testing.
    snippet3 = Markup(dedent('''
        <h1>Content of file_inclusion.html</h1>
        <pre>
        {examples:read_markup('file_inclusion.html')}
        </pre>
    '''))
    # Included HTML file should appear as is:
    result3 = dedent('''
        <h1>Content of file_inclusion.html</h1>
        <pre>
        <div>
            Sample <strong>inclusion</strong> file.
        </div>

        </pre>
    ''')

    # Example 4.
    # Same as example 3, but file name is passed via substitution.
    # Note that we need to add quotes around file name to make the invocation syncactically correct.
    filename = '"file_inclusion.html"'

    snippet4 = Markup(dedent('''
        <h1>Content of file_inclusion.html</h1>
        <pre>
        {examples:read_markup({filename})}
        </pre>
    '''))
    # Result is same as in example 3.
    result4 = result3

    # Example 5.
    # Same as example 4, use property for filename and alternative quotation.
    @property
    def filename_prop(self, context=None):
        return 'file_inclusion.html'  # no quotes here, they are moved to the template string, see the snippet:

    snippet5 = Markup(dedent('''
        <h1>Content of file_inclusion.html</h1>
        <pre>
        {examples:read_markup('{filename_prop}')}
        </pre>
    '''))
    # Result is same as in example 3 and 4.
    result5 = result4

class BadTemplate1(BaseTemplate):

    # file paths outside the root directory should fail
    snippet = Markup(dedent('''
        {examples:read('../core.py')}
    '''))

class BadTemplate2(BaseTemplate):

    # file paths outside the root directory should fail
    snippet = Markup(dedent('''
        {examples:read('foo/bar/../../../core.py')}
    '''))

class BadTemplate3(BaseTemplate):

    # simply bad path should cause bare Exception
    snippet = Markup(dedent('''
        {examples:read('/core.py')}
    '''))

def test_file_inclusion():
    template = GoodTemplate()
    context = template.render()
    print('--- sample1')
    print(context['snippet1'])
    print('--- sample2')
    print(context['snippet2'])
    assert context['snippet2'].strip() == template.result2.strip()
    print('--- sample3')
    print(context['snippet3'])
    assert context['snippet3'].strip() == template.result3.strip()
    assert context['snippet4'].strip() == template.result4.strip()
    assert context['snippet5'].strip() == template.result5.strip()

def test_vulnerabilities():
    template = BadTemplate1()
    try:
        context = template.render()
        print('--- clabate core:')
        print(context['snippet'])
        assert False, 'Hooray!'
    except BadPathError:
        print('Access blocked, that\'s good')

    template = BadTemplate2()
    try:
        context = template.render()
        print('--- clabate core:')
        print(context['snippet'])
        assert False, 'Hooray!'
    except BadPathError:
        print('Access blocked, that\'s good')

    template = BadTemplate3()
    try:
        context = template.render()
        print('--- clabate core:')
        print(context['snippet'])
        assert False, 'Hooray!'
    except Exception as e:
        assert e.__class__.__name__ == 'Exception'
        print('Bad path caused an exception %s, that\'s good' % e.__class__.__name__)
