'''
Invocation examples.

:copyright: Copyright 2019-2021 AXY axy@declassed.art
:license: BSD, see LICENSE for details.
'''

from textwrap import dedent

from ..markup import Invoke, Template

class GoodTemplate1(Template):

    '''
    Example 1.
    The idea.
    '''
    snippet1 = dedent('''
        The idea behind invoke {invoke_idea:is that format_spec is passed as is
        to the __format__ method where we can compile and run it}
    ''')

    class InvokeIdea:

        def __format__(self, format_spec):
            return format_spec

    invoke_idea = InvokeIdea()

    result1 = dedent('''
        The idea behind invoke is that format_spec is passed as is
        to the __format__ method where we can compile and run it
    ''')

    '''
    Example 2.
    Real example.
    '''
    snippet2 = dedent('''
        So Invoke class provides __format__ method that does the job.
        Anything after colon is evaluated as a pythonic expression.
        We need to subclass it and define our methods.
        {my_invoke:one()}
        {my_invoke:two()}
        {my_invoke:three}  (yes, this is an attribute of my_invoke)
    ''')

    class MyInvoke(Invoke):

        def one(self):
            return '1'

        def two(self):
            return '2'

        three = '3'

    my_invoke = MyInvoke()

    result2 = dedent('''
        So Invoke class provides __format__ method that does the job.
        Anything after colon is evaluated as a pythonic expression.
        We need to subclass it and define our methods.
        1
        2
        3  (yes, this is an attribute of my_invoke)
    ''')

    '''
    Example 3.
    '''
    snippet3 = dedent('''
        Parameters may contain any built-in types supported in python syntax (mind braces escaping):
        {arguments:show(
            'string',
            True, False,
            1, 2.0, 3.4,
            a_complex = 5.0+6j,
            a_dict = {{'foo': 'bar'}},
            a_set = {{'foo', 'bar'}},
            a_list = [1, 2, 3]
        )}
    ''')

    class Arguments(Invoke):

        def show(self, *args, **kwargs):
            return 'args: %s\nkwargs: %s' % (args, kwargs)

    arguments = Arguments()

    result3 = dedent('''
        Parameters may contain any built-in types supported in python syntax (mind braces escaping):
        args: ('string', True, False, 1, 2.0, 3.4)
        kwargs: {{'a_complex': (5+6j), 'a_dict': {{'foo': 'bar'}}, 'a_set': {{'foo', 'bar'}}, 'a_list': [1, 2, 3]}}
    ''')

class BadTemplate1(Template):

    '''
    Example 4.
    '''
    snippet = dedent('''
        Names should not start with underscore.
        This won't work: {i_want_this:_working()}
    ''')

    class IWantThis(Invoke):

        def _working(self):
            return 'It works!!!'

    i_want_this = IWantThis()

    error = 'Denied access to _working'

class BadTemplate2(Template):

    '''
    Example 5.
    '''
    snippet = dedent('''
        No function calls are allowed in params.
        This won't work: {i_want_this:working(tuple(1,2,3))}
    ''')

    class IWantThis(Invoke):

        def working(self, arg):
            return 'It works: %s!!!' % arg

    i_want_this = IWantThis()

    error = 'Calls are not allowed in parameters: working(tuple(1,2,3))'

def test_invoke():
    template = GoodTemplate1()
    context = template.render()
    assert context['snippet1'].strip() == template.result1.strip()
    assert context['snippet2'].strip() == template.result2.strip()
    # XXX: the following test is unstable because of random order of items in the set
    #assert context['snippet3'].strip() == context['result3'].strip()  # Use context['result3'] because of braces escaping

    template = BadTemplate1()
    try:
        context = template.render()
        print(context['snippet'])
        assert False, 'Hooray!'
    except Exception as e:
        assert e.args[0] == template.error

    template = BadTemplate2()
    try:
        context = template.render()
        print(context['snippet'])
        assert False, 'Hooray!'
    except Exception as e:
        assert e.args[0] == template.error
