'''
Markup examples.

:copyright: Copyright 2021 AXY axy@declassed.art
:license: BSD, see LICENSE for details.
'''

from textwrap import dedent
from types import SimpleNamespace

from ..markup import MarkupTemplate, Markup, TagAttr

'''
Someone may ask "Can I use Markup class from markupsafe instead of clabate Markup?
Why not using markupsafe from the very beginning?"

Well there's a couple of reasons to use my own Markup class.

First, I tried to avoid using third party packages from core modules.

Second, the Markup class from markupsafe lacks __format__ method that would return Markup.
It returns string which gets escaped by clabate.

But you can use markupsafe.Markup. At least, all examples below do work if you uncomment
the following lines:

import markupsafe

class Markup(markupsafe.Markup):
    def __format__(self, format_spec):
        return Markup(super().__format__(format_spec))

'''

class GoodTemplate1(MarkupTemplate):

    '''
    Example 1.
    HTML boilerplate.
    '''
    html = Markup(dedent('''
        <html>
        {head}
        {body}
        </html>
    '''))

    head = Markup(dedent('''
        <head>
            <title>{title}</title>
        </head>
    '''))

    title = 'Example'

    body = Markup(dedent('''
        <body>
            <header>
                {header}
            </header>
            <main>
                {main}
            </main>
            <footer>
                {footer}
            </footer>
        </body>
    '''))

    header = Markup(dedent('''
        This is <strong>header</strong>.
    '''))

    main = Markup(dedent('''
        Main content goes <strong>here</strong>.
    '''))

    footer = Markup(dedent('''
        This is <strong>footer</strong>.
    '''))

    result = dedent('''
        <html>

        <head>
            <title>Example</title>
        </head>


        <body>
            <header>

        This is <strong>header</strong>.

            </header>
            <main>

        Main content goes <strong>here</strong>.

            </main>
            <footer>

        This is <strong>footer</strong>.

            </footer>
        </body>

        </html>
    ''')

class GoodTemplate2(MarkupTemplate):

    '''
    Example 2.
    All values are escaped. Including complex data and missing values:
    '''
    attr1 = 'String attributes are <strong>escaped</strong>'
    snippet1 = dedent('''
        {attr1}
        {init_kwarg}
        {render_kwarg}
        What is <a.b[name].c[1].d>? It's {a.b[name].c[1].d}!
        And what is foo.bar? {foo.bar}
    ''')

    a = SimpleNamespace(
        b = dict(
            name = SimpleNamespace(
                c = [
                    0,
                    SimpleNamespace(
                        d = '<wow>'
                    )
                ]
            )
        )
    )
    result1 = dedent('''
        String attributes are &lt;strong&gt;escaped&lt;/strong&gt;
        __init__ kwargs are &lt;strong&gt;escaped&lt;/strong&gt;
        render kwargs are &lt;strong&gt;escaped&lt;/strong&gt;
        What is &lt;a.b[name].c[1].d&gt;? It's &lt;wow&gt;!
        And what is foo.bar? &lt;foo&gt; is missing from the rendering context!
    ''')

    '''
    And values are escaped only once!
    '''
    snippet2 = dedent('''
        Values are escaped only once:{snippet1}
        Otherwise we'd get &amp;
    ''')
    result2 = dedent('''
        Values are escaped only once:
        String attributes are &lt;strong&gt;escaped&lt;/strong&gt;
        __init__ kwargs are &lt;strong&gt;escaped&lt;/strong&gt;
        render kwargs are &lt;strong&gt;escaped&lt;/strong&gt;
        What is &lt;a.b[name].c[1].d&gt;? It's &lt;wow&gt;!
        And what is foo.bar? &lt;foo&gt; is missing from the rendering context!

        Otherwise we'd get &amp;amp;
    ''')

    def missing(self, name):
        return '<%s> is missing from the rendering context!' % name

class GoodTemplate3(MarkupTemplate):

    '''
    Example 3.
    Same as example 2, but use markup:
    '''
    attr1 = Markup('Markup attributes are <strong>unchanged</strong>')
    snippet1 = Markup(dedent('''
        {attr1}
        {init_kwarg}
        {render_kwarg}
        What is <a.b[name].c[1].d>? It's {a.b[name].c[1].d}!
        And what is foo.bar? {foo.bar}
    '''))

    a = SimpleNamespace(
        b = dict(
            name = SimpleNamespace(
                c = [
                    0,
                    SimpleNamespace(
                        d = Markup('<span style="color:green">wow</span>')
                    )
                ]
            )
        )
    )
    result1 = dedent('''
        Markup attributes are <strong>unchanged</strong>
        __init__ kwargs are <strong>unchanged</strong>
        render kwargs are <strong>unchanged</strong>
        What is <a.b[name].c[1].d>? It's <span style="color:green">wow</span>!
        And what is foo.bar? <span style="color:red">foo</span> is missing from the rendering context!
    ''')

    def missing(self, name):
        return Markup('<span style="color:red">%s</span> is missing from the rendering context!' % name)

    '''
    And values are escaped only once!
    '''
    snippet2 = Markup(dedent('''
        Values are escaped only once:{snippet1}
        Otherwise we'd get &amp; instead of bare &
    '''))
    result2 = dedent('''
        Values are escaped only once:
        Markup attributes are <strong>unchanged</strong>
        __init__ kwargs are <strong>unchanged</strong>
        render kwargs are <strong>unchanged</strong>
        What is <a.b[name].c[1].d>? It's <span style="color:green">wow</span>!
        And what is foo.bar? <span style="color:red">foo</span> is missing from the rendering context!

        Otherwise we'd get &amp; instead of bare &
    ''')

class GoodTemplate4(MarkupTemplate):

    '''
    Example 4.
    Tag attributes.
    '''
    snippet1 = Markup(dedent('''
        <span data-link="{link1}">wrong attribute escaping</span>
    '''))
    link1 = '<a href="https://declassed.art">declassed.art</a>'
    result1 = dedent('''
        <span data-link="&lt;a href="https://declassed.art"&gt;declassed.art&lt;/a&gt;">wrong attribute escaping</span>
    ''')

    snippet2 = Markup(dedent('''
        <span data-link={link2}>correct attribute escaping</span>
    '''))
    link2 = TagAttr('<a href="https://declassed.art">declassed.art</a>')
    result2 = dedent('''
        <span data-link='&lt;a href="https://declassed.art"&gt;declassed.art&lt;/a&gt;'>correct attribute escaping</span>
    ''')

    snippet3 = Markup(dedent('''
        <span data-link={escape:attr('<a href="https://declassed.art">declassed.art</a>')}>inline attribute escaping</span>
    '''))
    result3 = dedent('''
        <span data-link='&lt;a href="https://declassed.art"&gt;declassed.art&lt;/a&gt;'>inline attribute escaping</span>
    ''')

    snippet4 = Markup(dedent('''
        Inline javascript attribute escaping
        <p onclick={escape:attr("""
            run("<script>this()</script>");
            run('<script>that()</script>');
        """)}>Click me!</p>
    '''))
    result4 = dedent('''
        Inline javascript attribute escaping
        <p onclick="run(&quot;&lt;script&gt;this()&lt;/script&gt;&quot;);&#10;    run('&lt;script&gt;that()&lt;/script&gt;');">Click me!</p>
    ''')

class GoodTemplate5(MarkupTemplate):

    '''
    Example 5.
    Markup inside non-markup.
    Sometimes it's necessary to insert some markup in a text to be escaped. Useful for wiki-alike systems.
    '''
    snippet1 = dedent('''
        When a text is going to be <escaped> but you need
        to insert {escape:markup("""<span style="color:red">your custom HTML</span>""")}
    ''')
    result1 = dedent('''
        When a text is going to be &lt;escaped&gt; but you need
        to insert <span style="color:red">your custom HTML</span>
    ''')

class GoodTemplate6(MarkupTemplate):

    '''
    Example 6.
    Format specification is escaped because of recursive nature of `string.Formatter.vformat()` method.
    '''
    snippet1 = dedent('''
        This is a <literal text> preceeding the {substitution: And this is a <literal text>
        for {nested_substitution: which gets <doubly escaped> if not a markup!}}
    ''')
    class MarkupSubst:
        def __format__(self, format_spec):
            return Markup(format_spec)
    class PlainSubst:
        def __format__(self, format_spec):
            return format_spec
    substitution = MarkupSubst()
    nested_substitution = PlainSubst()
    result1 = dedent('''
        This is a &lt;literal text&gt; preceeding the  And this is a &lt;literal text&gt;
        for  which gets &amp;lt;doubly escaped&amp;gt; if not a markup!
    ''')
    # So, it's better use Markup for such format strings in MarkupTemplate.

def test_markup():
    template = GoodTemplate1()
    context = template.render()
    # dedent does not strip inner empty lines:
    assert ('\n'.join(s.rstrip() for s in context['html'].splitlines())).strip() == template.result.strip()

    template = GoodTemplate2(init_kwarg='__init__ kwargs are <strong>escaped</strong>')
    context = template.render(render_kwarg='render kwargs are <strong>escaped</strong>')
    assert context['snippet1'].strip() == template.result1.strip()
    assert context['snippet2'].strip() == template.result2.strip()

    template = GoodTemplate3(init_kwarg=Markup('__init__ kwargs are <strong>unchanged</strong>'))
    context = template.render(render_kwarg=Markup('render kwargs are <strong>unchanged</strong>'))
    assert context['snippet1'].strip() == template.result1.strip()
    assert context['snippet2'].strip() == template.result2.strip()

    template = GoodTemplate4()
    context = template.render()
    assert context['snippet1'].strip() == template.result1.strip()
    assert context['snippet2'].strip() == template.result2.strip()
    assert context['snippet3'].strip() == template.result3.strip()
    assert context['snippet4'].strip() == template.result4.strip()

    template = GoodTemplate5()
    context = template.render()
    assert context['snippet1'].strip() == template.result1.strip()

    template = GoodTemplate6()
    context = template.render()
    assert context['snippet1'].strip() == template.result1.strip()
