'''
MarkupTemplate, a template for XML/HTML markup.
Escaping and minification helpers.

:copyright: Copyright 2019-2021 AXY axy@declassed.art
:license: BSD, see LICENSE for details.
'''

import re
from xml.sax import saxutils

from .core import Template
from .invoke import Invoke

# All dependencies are optional. Do not minify if a particular minifier is not installed.

# CSS and JS minification is performed by rcssmin and rjsmin.
# For HTML I haven't found anything fast and simple yet -- Axy.

try:
    import rjsmin
    _rjsmin_available = True
except ImportError:
    _rjsmin_available = False

try:
    import rcssmin
    _rcssmin_available = True
except ImportError:
    _rcssmin_available = False

class Escape(Invoke):
    '''
    This is a double-purpose class.
    First, as a callable it is intended for overriding ``Template.escape`` in ``MarkupTemplate``.
    Second, it is inherited from Invoke and contains methods that can be used from templates.

    For now it provides ``attr`` method for templates:

        <p onclick={escape:attr("""
            do_this();
            do_that();
        """)}>Click me!</p>

    Yes, multiline attributes are okay!
    '''
    def attr(self, text, strip=True):
        '''
        Escape tag attribute.

        Args:
            text: attribute value
            strip: strip leading and trailing spaces (default: True)

        Returns:
            an instance of TagAttr object
        '''
        if strip:
            text = text.strip()
        return TagAttr(text)

    def markup(self, text):
        '''
        Do not escape text.

        This method is intended to keep markup in a non-markup string.

        Args:
            text: contains markup

        Returns:
            an instance of Markup object
        '''
        # `text`, which actually is a part of format specification, is already escaped
        # because of recursive nature of `string.Formatter.vformat()` method.
        # That's why we need to unescape it.
        return Markup(saxutils.unescape(text))

    def __call__(self, value):
        '''
        Escape value.

        Value can be a bare string or an instance of some markup object which provides ``__html__`` method.

        Args:
            value: value to escape

        Returns:
            string if value was bare string or the result of ``value.__html__()``
        '''
        if callable(getattr(value, '__html__', None)):
            # Return markup as is.
            return value.__html__()
        else:
            # Escape strings.
            return saxutils.escape(value)

class MarkupTemplate(Template):
    '''
    This subclass implements XML escaping and supports __html__ method for values.
    '''
    escape = Escape()

class Markup(str):
    '''
    Markup wrapper for raw XML/HTML code.
    '''

    def __html__(self):
        return self

    def __format__(self, format_spec):
        '''
        Override formatting method to return Markup.
        This is necessary for proper escaping of substitutions.
        '''
        return Markup(super().__format__(format_spec))

    # Override other formatting methods, just for completeness.

    def format(self, *args, **kwargs):
        return Markup(super().format(*args, **kwargs))

    def format_map(self, mapping):
        return Markup(super().format_map(mapping))

class TagAttr(Markup):
    '''
    Escape tag attribute value.

    Alternatively, use {escape:attr('...')}
    '''
    def __html__(self):
        return saxutils.quoteattr(self)

class MinifiedMarkupTemplate(MarkupTemplate):
    '''
    Remove comments and unnecessary spaces from resulting markup.
    Preserve spaces in <pre> tags.

    An alernative to this subclass is ``minified`` function.
    '''
    def post_process(self, formatted_value):
        return minified(formatted_value)

_re_html_comments = re.compile('<!--.*?-->', flags=re.DOTALL)
_re_special_tags = re.compile(r'<(script|style|pre)(>| [^>]*>)(.*)?</\1>', flags=re.DOTALL)

def minified(markup):
    '''
    Remove comments and unnecessary spaces from resulting markup.
    Preserve spaces in <pre> tags.

        class MyTemplate(clabate.MarkupTemplate):

            html = minified("""
                <html>
                ...
                </html>
            """)

    An alernative to this function is ``MinifiedMarkupTemplate``.
    '''
    # remove HTML comments
    markup = _re_html_comments.sub('', markup)
    # The following tags need special processing:
    # * script: use rjsmin
    # * style: use rcssmin
    # * pre: preserve spaces
    result = []
    start_index = 0
    for matchobj in _re_special_tags.finditer(markup):
        # process leading markup
        result.append(_minify_markup(markup[start_index:matchobj.start()]))
        # process opening tag: remove spaces
        result.append(
            _re_spaces.sub(' ', markup[matchobj.start():matchobj.start(3)])
        )
        # process inner content
        tag = matchobj.group(1)
        if tag == 'script':
            content = _minify_js(matchobj.group(3))
        elif tag == 'style':
            content = _minify_css(matchobj.group(3))
        else:
            content = matchobj.group(3)
        result.append(content)
        # append closing tag
        result.append(markup[matchobj.end(3):matchobj.end()])
        # bump start_index
        start_index = matchobj.end()
    # process ending markup
    result.append(_minify_markup(markup[start_index:]))
    return Markup(''.join(result))

_re_spaces = re.compile(r'\s+')
_re_intertag_spaces = re.compile(r'>\s+<')

def _minify_markup(markup):
    # Simply remove unnecessary spaces.
    markup = _re_spaces.sub(' ', markup)
    markup = _re_intertag_spaces.sub('><', markup)
    return markup.strip()

def _minify_js(js):
    if _rjsmin_available:
        return rjsmin.jsmin(js)
    else:
        return js

def _minify_css(css):
    if _rcssmin_available:
        return rcssmin.cssmin(css)
    else:
        return css
